/******************************************************************************/
/*   amcc.c -- AMCC S5920 PCI Target Interface driver                         */
/*   (c) 2001 Greg Burley (burley@ociw.edu)                                   */
/*            Drew Moore  (acm3035@rit.edu)                                   */
/*                                                                            */
/*   Much of this driver is shamelessly based on the code for the             */
/*   AMCC S5933 PCI Matchmaker driver                                         */
/*   (c) 1997 Andrea Cisternino  (acister@pcape1.pi.infn.it)                  */
/*                                                                            */
/*   This program is free software; you can redistribute it and/or            */
/*   modify it under the terms of the GNU General Public License as           */
/*   published by the Free Software Foundation (www.gnu.org)                  */
/*                                                                            */
/******************************************************************************/

/*------------------------------------------------------------------------------
 *
 *  Driver main file -- this file contains the driver code.
 *
 *----------------------------------------------------------------------------*/

#ifndef __KERNEL__
# define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/config.h>               /* for CONFIG_* + CONFIG_PCI */
#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#   define MODVERSIONS                  /* force it on */
#endif

#ifdef MODVERSIONS
#   include <linux/modversions.h>
#endif

#include <linux/module.h>

#include <linux/sched.h>                /* current and everything */
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/stat.h>
#include <linux/delay.h>                /* for <asm/delay.h> */
#include <asm/io.h>                     /* for inb(), outb() etc. */
#include <asm/system.h>                 /* cli(), *_flags */
#include <asm/pgtable.h>
#include <asm/uaccess.h>

#include "amcc.h"                       /* S5920 definitions */

#ifndef AMCC_DEBUG
    EXPORT_NO_SYMBOLS;              /* otherwise, leave global symbols visible */
#endif
MODULE_LICENSE("GPL");
/*-- Defines ---------------------------------------------------------*/

#define AMCC_NAME               "amcc"
#define AMCC_MAGIC              0x35393230  /* "5920" */
#define AMCC_RW_TIMEOUT         (HZ*5)      /* 5 sec. */
#define AMCC_TIMEOUT            50          /* timeout in microseconds */
#define AMCC_MAX_RETRYS         1000

/* lock bits for concurrent access */
#define AMCC_P2A_LOCK_BIT       0
#define AMCC_A2P_LOCK_BIT       1
#define AMCC_TEST_IRQ_BIT       2

/* Miscellaneous */
#define AMCC_POLL_DELAY         20UL	/* in usecs */

/* return codes for interrupts */
#define AMCC_IRQ_READ_OK        1
#define AMCC_IRQ_WRITE_OK       2
#define AMCC_IRQ_MASTER_ABORT   3
#define AMCC_IRQ_TARGET_ABORT   4
#define AMCC_IRQ_OMB_EMPTY      5
#define AMCC_IRQ_IMB_FULL       6

#define PCI_DEVICE_ID_AMCC_S5920    0x5920

#define AMCC_MAJOR              125     /* Major device number requested */
#define AMCC_FIFO_SIZE          16384   /* FIFO is 16384 16-bit words */
#define P2A_BUFFER_SIZE         4096
#define A2P_BUFFER_SIZE         65536

#define MIN(a,b)                (((a) < (b)) ? (a) : (b))

/*-- Structures ------------------------------------------------------*/

/*  Pass-thru info -- holds information about a single pass-thru region. */
struct s5920_pt {
    u_int       size;           /* dimensions */
    u_int       phys_addr;      /* real address */
    char       *virt_addr;      /* after ioremap call */
};


/* Hardware info -- holds global data for a single PCI board. */
struct amcc_s5920 {
    struct pci_dev *pcidev;
    int         major;
    u_long      op_regs;        /* this is BADDR[0] */
    int         op_regs_type;   /* MEM or I/O operation registers */
    u_char      irq;            /* interrupt number */
    u_char      ptnum;          /* number of pass-thru regions */
    struct s5920_pt pts[5];     /* pass-thru regions */
    int         timeout;        /* timeout for async waits */
};

/* Buffer info -- holds data for Pass-Thru buffer handling. */
struct amcc_buffer {
    volatile int   irq_code;
    volatile u_int copy_size;
    volatile u_int bytes_done;
    volatile u_int size;
    u_int  async;
    char   *buf;
};



/*------------------------------------------------------------------------------
 *
 *  Static variables.
 *  These can be seen by every process using the device.
 *
 *----------------------------------------------------------------------------*/

struct amcc_s5920 amcc_dev;         /* General data */

struct amcc_buffer p2a;             /* P2A buffer control struct */
struct amcc_buffer a2p;             /* A2P buffer control struct */

static u_long lock = 0UL;           /* generic lock dword */

static u32  badrs[6] =
{
    PCI_BASE_ADDRESS_0,             /* PCI bus operations register */
    PCI_BASE_ADDRESS_1,
    PCI_BASE_ADDRESS_2,
    PCI_BASE_ADDRESS_3,
    PCI_BASE_ADDRESS_4,
    PCI_BASE_ADDRESS_5,             /* not used in S5920 */
};


/*------------------------------------------------------------------------------
 *
 *  Flags.
 *  These variables can be overridden at load time by the insmod cmd.
 *
 *----------------------------------------------------------------------------*/

int vid = PCI_VENDOR_ID_AMCC;
int did = PCI_DEVICE_ID_AMCC_S5920;

/*------------------------------------------------------------------------------
 *
 *  Function prototypes
 *
 *----------------------------------------------------------------------------*/

int init_module (void);
void cleanup_module (void);

static ssize_t amcc_read (struct file *, char *, size_t, loff_t *);
static ssize_t amcc_write (struct file *, const char *, size_t, loff_t *);
static int amcc_ioctl (struct inode *, struct file *, unsigned int, unsigned long);
static int amcc_open (struct inode *, struct file *);
static int amcc_release (struct inode *, struct file *);

static int  amcc_init_dev (struct amcc_s5920 *, struct pci_dev *);
static void amcc_free_dev (void);

/* static void amcc_irq_handler (int, void *, struct pt_regs *); */



/*------------------------------------------------------------------------------
 *
 *  Linux kernel main entry point.
 *  For a definition of its fields see <linux/fs.h>.
 *
 *----------------------------------------------------------------------------*/

static struct file_operations amcc_fops =
{
    open:       amcc_open,
    release:    amcc_release,
    read:       amcc_read,
    write:      amcc_write,
    ioctl:      amcc_ioctl,
};

/*-----------------------------------------------------------------------------
 *
 *  Inline Functions
 *
 *----------------------------------------------------------------------------*/

extern inline __u32
amcc_read_opreg (int reg)
{
    extern struct amcc_s5920 amcc_dev;

    if (amcc_dev.op_regs_type == PCI_BASE_ADDRESS_SPACE_IO)
        return ((__u32) inl ((unsigned short) (amcc_dev.op_regs + reg)));
    else
        return (readl ((amcc_dev.op_regs + reg)));
}

extern inline void
amcc_write_opreg (__u32 value, int reg)
{
    extern struct amcc_s5920 amcc_dev;

    if (amcc_dev.op_regs_type == PCI_BASE_ADDRESS_SPACE_IO)
        outl (value, (unsigned short) (amcc_dev.op_regs + reg));
    else
        writel (value, (amcc_dev.op_regs + reg));
}


/*------------------------------------------------------------------------------
 *
 *  Standard module initialization function.
 *  This function scans the PCI bus looking for the right board. Support
 *  functions can be found in file support.c
 *
 *----------------------------------------------------------------------------*/

int
init_module (void)
{
    extern struct amcc_s5920 amcc_dev;

    extern struct amcc_buffer p2a;
    extern struct amcc_buffer a2p;

    struct pci_dev *pcidev = NULL;
    int  result = 0L;

    printk (KERN_DEBUG "amcc_init >>> START <<<\n");

    memset (&amcc_dev, 0, sizeof (amcc_dev));
    memset (&p2a, 0, sizeof (p2a));
    memset (&a2p, 0, sizeof (a2p));

    if (!pci_present())
        return(-ENODEV);

    if ((pcidev = pci_find_device((u_short)vid, (u_short)did, pcidev))){

        result = register_chrdev(AMCC_MAJOR, AMCC_NAME, &amcc_fops);
        if (result == 0)
            amcc_dev.major = AMCC_MAJOR;
        else
            amcc_dev.major = result;
        printk (KERN_WARNING "amcc_dev.major = %d\n", amcc_dev.major);

        if (amcc_dev.major == -EBUSY){
            printk (KERN_WARNING "amcc: unable to get major number\n");
            return (-EIO);
        }

        amcc_dev.pcidev = pcidev;
        amcc_dev.timeout = AMCC_RW_TIMEOUT;
        result = amcc_init_dev(&amcc_dev, pcidev);
    }

    printk (KERN_INFO "amcc: AMCC S5920 Rev 1.2 for 2.4.x kernel \n");
    return (result);
}

/*------------------------------------------------------------------------------
 *
 *  Standard module release function.
 *
 *----------------------------------------------------------------------------*/

void
cleanup_module (void)
{
    amcc_free_dev ();
    printk (KERN_DEBUG "amcc: >>> STOP <<<\n");
}

/*------------------------------------------------------------------------------
 * 
 *  Standard open() entry point.
 *  It simply increments the module usage count.
 *
 *----------------------------------------------------------------------------*/

static int
amcc_open (struct inode *inode, struct file *file)
{
    printk (KERN_DEBUG "amcc: open\n");

    /* Get the minor device number in case there is more than one  *
     * physical device using the driver                            */

    printk (KERN_DEBUG "Device minor number %d.%d\n",
        inode->i_rdev >>8, inode->i_rdev & 0xff);

    MOD_INC_USE_COUNT;
    return (0);
}

/*------------------------------------------------------------------------------
 *
 *  Standard release() entry point.
 *  This function is called by the close() system call.
 *
 *----------------------------------------------------------------------------*/

static int
amcc_release (struct inode *inode, struct file *file)
{
    printk (KERN_DEBUG "amcc: close\n");
    MOD_DEC_USE_COUNT;

    return(0);
}

/*------------------------------------------------------------------------------
 *
 *  Standard read() entry point.
 *  It is used to read from the Add-on bus to PCI bus.
 *  It fills a kernel buffer with the specified number of bytes.
 *
 *  Blocking + Non-Blocking I/O options need to be worked on.
 *  The read() code is hardware specific.
 *
 *----------------------------------------------------------------------------*/

static ssize_t
amcc_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
    extern struct amcc_s5920 amcc_dev;
    extern struct amcc_buffer a2p;
    ssize_t  total_bytes_read = 0;
    u32 *ptr, *pcnt, r;
    u16 *rbuf;
    int j, k, nwords, retrys;

    r=0;
    k=0;
    j=0;
    retrys=0;
    if (count == 0)
        return (0);

    /* Is the Pass_Thru + FIFO available? */
    if (test_and_set_bit (AMCC_A2P_LOCK_BIT, &lock)) {
        printk (KERN_DEBUG "amcc: Read PT+FIFO already locked\n");
        return (-EBUSY);
    }

    /* Set the async flag correctly -- we don't do much with this yet. *
     * Eventually we will implement blocking + non-blocking mode.      */
    if (file->f_flags & O_NONBLOCK){
        a2p.async = 1L;
    }
    else {
        a2p.async = 0L;
    }

    /* Set up to read the device into the a2p buffer */
    ptr = (u32 *)(amcc_dev.pts[1]).virt_addr;
    pcnt = (u32 *)(amcc_dev.pts[2]).virt_addr;
    rbuf = (u16 *)a2p.buf;
    nwords = 0;

    a2p.bytes_done = 0;
    a2p.copy_size = MIN (count, a2p.size);

#define AMCC_ROUTINE   3

#if (AMCC_ROUTINE==0)
/*  Routine 1 -- Read until done, or until the FIFO goes empty. */
/*  If the FIFO is not ready, return immediately.               */
    for (j=0; j<a2p.copy_size/2; j++) {
        if (!(amcc_read_opreg (AMCC_OP_REG_IMB) & IMB_RX_RDY))
            break;
        rbuf[j] = (u16) readl ((u32 *)(&ptr[j]));
        a2p.bytes_done +=2;
    }
#endif

#if (AMCC_ROUTINE==1)
/*  Routine 1 -- Read until done, or until the FIFO goes empty. */
    j = 0;
    while (a2p.bytes_done < a2p.copy_size  && retrys < AMCC_MAX_RETRYS) {
        if (amcc_read_opreg (AMCC_OP_REG_IMB) & IMB_RX_RDY){
            rbuf[j] = (u16) readl ((u32 *)(&ptr[j]));
            a2p.bytes_done +=2;
            retrys=0;
            j++;
        }
        else {
            if (a2p.copy_size==2)break;
            retrys += 1;
            udelay(10);
        }
    }
#endif

#if (AMCC_ROUTINE==2)
    /* Routine 2 -- Check the FIFO word counter first.  Then,  */
    /* read nwords from FIFO.                                  */
    while (a2p.bytes_done < a2p.copy_size) {
        if ( !(amcc_read_opreg(AMCC_OP_REG_IMB) & IMB_RX_RDY))
            break;
        nwords = (u32) readl ((u32 *)(pcnt) );
        for (j=0; j < nwords; j++){
            rbuf[j] = (u16) readl ((u32 *)(&ptr[j]) );
            a2p.bytes_done +=2;
        }
    }
#endif

#if (AMCC_ROUTINE==3)
    j = 0;
    while (a2p.bytes_done < a2p.copy_size  && retrys < AMCC_MAX_RETRYS) {
        r = amcc_read_opreg (AMCC_OP_REG_IMB);
        if (r & IMB_RX_RDY){
            if (a2p.copy_size-a2p.bytes_done > 255 && (r & IMB_RX_PAE) ){
                for (k=0; k<127; k++){
                    rbuf[j] = (u16) readl ((u32 *)(&ptr[k]));
                    j++;
                    a2p.bytes_done += 2;
                }
            }
            else {
                rbuf[j] = (u16) readl ((u32 *)(&ptr[0]));
                a2p.bytes_done +=2;
                j++;
            }
            retrys=0;
        }
        else {
            if (a2p.copy_size==2)break;
            retrys += 1;
            udelay(10);
        }
    }
#endif

#if (AMCC_ROUTINE==4)
    retrys=0;
    for (j=0; j<a2p.copy_size/2; j++) {
        r = amcc_read_opreg (AMCC_OP_REG_IMB);
        if (!(r & IMB_RX_RDY)) break;
        if (a2p.copy_size-a2p.bytes_done > 255 && (r & IMB_RX_PAE) ){
            amcc_write_opreg ((u32) 0x00000008, AMCC_OP_REG_PTCR);
            for (k=0; k<127; k++){
                rbuf[j] = (u16) readl ((u32 *)(&ptr[k]));
                j++;
                a2p.bytes_done += 2;
            }
            j--;
            amcc_write_opreg ((u32) 0x00000000, AMCC_OP_REG_PTCR);
        }
        else {
            rbuf[j] = (u16) readl ((u32 *)(&ptr[0]));
            a2p.bytes_done +=2;
        }
    }

#endif


    if(copy_to_user(buf, a2p.buf, a2p.bytes_done))
        return(-EFAULT);
    total_bytes_read += a2p.bytes_done;
    count -= a2p.bytes_done;

    clear_bit (AMCC_A2P_LOCK_BIT, &lock);
    return (total_bytes_read);
}

/*------------------------------------------------------------------------------
 *
 *  Standard write() entry point.
 *  It is used to write to the PCI-to-Add-on FIFO. It copies the user
 *  buffer into our (smaller) kernel buffer and then performs a transfer.
 *  The write() code is hardware specific.
 *----------------------------------------------------------------------------*/
static ssize_t
amcc_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    extern struct amcc_s5920 amcc_dev;
    extern struct amcc_buffer p2a;
    ssize_t total_bytes_written = 0;
    u32 *ptr;
    u16 *wbuf;
    int j,k;

    /* Is the Pass-Thru + FIFO available? */
    if (test_and_set_bit (AMCC_P2A_LOCK_BIT, &lock)) {
        printk (KERN_DEBUG "amcc_write: Pass_Thru + FIFO already locked\n");
        return (-EBUSY);
    }

    /* Set the async flag correctly -- we don't do much with this yet. *
     * Someday we will implement blocking + non-blocking mode.         */
    if (file->f_flags & O_NONBLOCK)
        p2a.async = 1L;
    else
        p2a.async = 0L;

    /* Can we check to see if the Pass-Thru + FIFO is empty? */

    /* Set up to write to the device from the p2a buffer */
    ptr = (u32 *)(amcc_dev.pts[1]).virt_addr;
    wbuf = (u16 *)p2a.buf;

    /* Start write -- after each write, wait for EPLD to transmit the data  */
    /* ie wait for hardware flag to set, then check flag before each write  */
    /* to prevent overwriting the previous data as it is being transmitted  */
    while (count > 0){
        p2a.bytes_done = 0;
        p2a.copy_size = MIN (count, p2a.size);
        if (copy_from_user(p2a.buf, buf, p2a.copy_size))
            return (-EFAULT);
        for (j=0; j<p2a.copy_size/2; j++){
            for (k=0; k < AMCC_TIMEOUT; k++){
                udelay(1);
                if ( !(amcc_read_opreg(AMCC_OP_REG_IMB) & IMB_TX_BUSY))
                    break;
            }
            writel ( (u32)wbuf[j], (u32 *)(&ptr[j]) );
            udelay(5);
            p2a.bytes_done += 2;
        }
        count -= p2a.bytes_done;
        total_bytes_written += p2a.bytes_done;
    }
    clear_bit (AMCC_P2A_LOCK_BIT, &lock);
    return (total_bytes_written);
}


/*------------------------------------------------------------------------------
 *
 *  Standard ioctl() entry point.
 *
 *----------------------------------------------------------------------------*/
static int
amcc_ioctl (struct inode *inode, struct file *file,
    unsigned int cmd, unsigned long arg)
{
    extern struct amcc_s5920 amcc_dev;
    extern u_long lock;
    int ret = 0L;
    int j;

    switch (cmd) {
    case AMCC_RESET:            /* reset the whole board */
        amcc_write_opreg (RCR_RESET_MBFLAGS, AMCC_OP_REG_RCR);
        amcc_write_opreg (RCR_RESET_ADDON, AMCC_OP_REG_RCR);
        amcc_write_opreg ((u32) 0, AMCC_OP_REG_RCR);
        lock = 0L;
        break;

    case AMCC_RESET_ADDON:      /* reset the Add-on bus */
        amcc_write_opreg (RCR_RESET_ADDON, AMCC_OP_REG_RCR);
        for (j=0; j<100; j++)udelay(1000);
        amcc_write_opreg ((u32) 0, AMCC_OP_REG_RCR);
        lock = 0L;
        break;

    case AMCC_RESET_MBFLAGS:    /* reset the maibox flags */
        amcc_write_opreg (RCR_RESET_MBFLAGS, AMCC_OP_REG_RCR);
        break;

    case AMCC_RESET_FIFO:       /* reset the PCI card external FIFO */
        {
        u32  *addr = (u32 *) arg;
        u32 *ptr;
        u32  count=0;

        ptr = (u32 *)(amcc_dev.pts[1]).virt_addr;

        for (j=0; j<4*AMCC_FIFO_SIZE; j++) {
            if (!(amcc_read_opreg (AMCC_OP_REG_IMB) & IMB_RX_RDY))
                break;
            (void)readl((u32 *)(ptr));
            count +=2;
        }
        ret = put_user (count, addr);
        }
        break;

    case AMCC_GET_ICSR:         /* read ICSR register */
        {
        u32  *addr = (u32 *) arg;

        ret = put_user (amcc_read_opreg (AMCC_OP_REG_ICSR), addr);
        }
        break;

    case AMCC_GET_RCR:          /* read RCR register */
        {
        u32  *addr = (u32 *) arg;

        ret = put_user (amcc_read_opreg (AMCC_OP_REG_RCR), addr);
        }
        break;


    case AMCC_GET_PTCR:         /* read PTCR register */
        {
        u32  *addr = (u32 *) arg;

        ret = put_user (amcc_read_opreg (AMCC_OP_REG_PTCR), addr);
        }
        break;

    case AMCC_GET_MBEF:         /* read mailbox MBEF register */
        {
        u32  *addr = (u32 *) arg;

        ret = put_user (amcc_read_opreg (AMCC_OP_REG_MBEF), addr);
        if (ret == -EFAULT) return (ret);
        }
        break;


    case AMCC_SET_PTCR:         /* write to PTCR register */
        {

        u32  ptcr;

        ret = get_user (ptcr, (u32 *) arg);
        if (ret == -EFAULT)
            return (ret);
        amcc_write_opreg (ptcr, AMCC_OP_REG_PTCR);
        }
        break;

    case AMCC_SET_P2ABUF:       /* change P2A buffer size */
        ret = -EINVAL;
        break;

    case AMCC_SET_A2PBUF:       /* change A2P buffer size */
        ret = -EINVAL;
        break;

    case AMCC_WRITE_MB:         /* write to mailbox */
        ret = -EINVAL;
        break;

    case AMCC_READ_MB:          /* read from mailbox */
        ret = -EINVAL;
        break;

    case AMCC_SET_OMBIRQ:       /* set IRQ on OMB empty */
        ret = -EINVAL;
        break;

    case AMCC_SET_IMBIRQ:       /* set IRQ on IMB full */
        ret = -EINVAL;
        break;

    case AMCC_CLR_MBIRQ:        /* clear MB IRQ flags */
        ret = -EINVAL;
        break;

    case AMCC_TEST:             /* miscellaneous tests */
        {
        u32 *ptr, *pcnt, tmp, imb;
        u32 buf[1];

        buf[0] = 0xaabb5522;
        tmp = 0;

        ptr = (u32 *)(amcc_dev.pts[1]).virt_addr;
        pcnt = (u32 *)(amcc_dev.pts[2]).virt_addr;

        imb = amcc_read_opreg (AMCC_OP_REG_IMB);
        printk (KERN_DEBUG "amcc_tmp  imb %08x\n ", imb);
        printk (KERN_DEBUG "amcc_tmp  rxf %08x\n ", (IMB_TX_BUSY & imb));

        writel ( (u32)buf[0], (u32 *)(ptr));
        udelay(1);

        imb = amcc_read_opreg (AMCC_OP_REG_IMB);
        printk (KERN_DEBUG "amcc_tmp  imb %08x\n ", imb);
        printk (KERN_DEBUG "amcc_tmp  rxf %08x\n ", (IMB_TX_BUSY & imb));

        }
        break;

    default:
        ret = -EINVAL;
        break;
    }

    return (ret);
}


/*------------------------------------------------------------------------------
 *
 *  Standard mmap() entry point.
 *  Very simple for the moment.
 *
 *----------------------------------------------------------------------------*/

/*
static int
amcc_mmap (struct inode *inode, struct file *file, struct vm_area_struct *vma)
{
    return (0);
}
*/

/*------------------------------------------------------------------------------
 *
 *  Device initialization.
 *  This function initializes the board. It requests an interrupt and
 *  reserves an IO region for the operation registers. The Major number
 *  has already been allocated by the main init function.
 *
 *----------------------------------------------------------------------------*/

static int
amcc_init_dev (struct amcc_s5920 *dev, struct pci_dev *pcidev)
{

    extern struct amcc_buffer p2a;
    extern struct amcc_buffer a2p;
    u_long  ret;
    int  count;
    int j;

    /* Buffer allocation */

    p2a.buf = (char *) kmalloc(P2A_BUFFER_SIZE, GFP_KERNEL);
    p2a.size = P2A_BUFFER_SIZE;

    a2p.buf = (char *) kmalloc(A2P_BUFFER_SIZE, GFP_KERNEL);
    a2p.size = A2P_BUFFER_SIZE;
    
    if ((p2a.buf == NULL) || (a2p.buf == NULL)) {
        printk (KERN_WARNING "amcc: no memory for Pass-Thru buffers\n");
        amcc_free_dev ();
        return (-ENOMEM);
    }

    /* The following section retrieves from PCI_BASE_ADDRESS_0 the address
     * of the S5920 operations registers. It then registers the IO address
     * range with the system (ie 32 DWORD).                             */

    printk (KERN_DEBUG "amcc: pcidev_resource[0].start = 0x%08lx\n",
        pcidev->resource[0].start);

    /* Check for I/O or Memory Mapped Operation Registers */

    if (pcidev->resource[0].flags & IORESOURCE_IO){
        dev->op_regs=(pcidev->resource[0].start & PCI_BASE_ADDRESS_IO_MASK);
        ret = check_region (dev->op_regs, AMCC_OP_REG_SIZE);
        if (ret == -EBUSY) {
            printk (KERN_WARNING "amcc: I/O space already requested");
            printk (KERN_WARNING "at 0x%04lx\n", dev->op_regs);
            dev->op_regs = 0;
            amcc_free_dev ();
            return (-EIO);
        }
        request_region (dev->op_regs, AMCC_OP_REG_SIZE, AMCC_NAME);
        dev->op_regs_type = PCI_BASE_ADDRESS_SPACE_IO;
        printk (KERN_DEBUG "amcc: IO_MAP BADR[0] = 0x%08lx\n",dev->op_regs);
    }
    else {
        dev->op_regs=(pcidev->resource[0].start & PCI_BASE_ADDRESS_MEM_MASK);
        dev->op_regs = (u_long) ioremap (dev->op_regs, AMCC_OP_REG_SIZE);
        dev->op_regs_type = PCI_BASE_ADDRESS_SPACE_MEMORY;
        printk (KERN_DEBUG "amcc: MEM_MAP BADR[0] = 0x%08lx\n",dev->op_regs);
    }


    /* For each Pass-Thru region, extract information from the PCI base
     * addresses registers and remap each area into memory space.
     * As per the PCI spec, as soon as one base address is unused
     * (ie set to 0) any subsequent address is also unused.          */

    printk (KERN_DEBUG "amcc: Getting configuration for PT regions\n");

    for (count = 1; count <=4; count++){
        printk(KERN_DEBUG "amcc: PT[%d] base_address[%08lx] \n",
            count, pcidev->resource[count].start);
    }

    dev->ptnum = 0;
    for (count = 1; count <= 4; count++) {
        u32  val;
        u32  mask;

        if (pcidev->resource[count].start == 0) {
            printk(KERN_DEBUG "amcc: PT[%d]  blank -- not mapped\n",count);
            break;
        }

        if (pcidev->resource[count].start & PCI_BASE_ADDRESS_SPACE_IO) {
            printk(KERN_DEBUG "amcc: We only support memory mapped PT \n");
            continue;
        }

        cli ();
        pci_read_config_dword(pcidev, badrs[count], &val);
        pci_write_config_dword(pcidev, badrs[count], ~0);
        pci_read_config_dword(pcidev, badrs[count], &mask);
        pci_write_config_dword(pcidev, badrs[count], val);
        sti();

        printk (KERN_DEBUG "amcc: pt[%d]  mask %08x  val  %08x\n",
            count, mask, val);

        mask &= PCI_BASE_ADDRESS_MEM_MASK;

        dev->ptnum++;

        (dev->pts[count]).size = (~mask) + 1;
        (dev->pts[count]).phys_addr =
            (pcidev->resource[count].start & PCI_BASE_ADDRESS_MEM_MASK);
        (dev->pts[count]).virt_addr =
            ioremap ((dev->pts[count]).phys_addr,(dev->pts[count]).size);

        printk (KERN_DEBUG "amcc: PT[%d]  Size %8d  Phys %08x  Virt %08x\n",
            (count),
            (dev->pts[count]).size,
            (dev->pts[count]).phys_addr,
            (int)(dev->pts[count]).virt_addr);
    }

    /* Board reset -- set some defaults */
    amcc_write_opreg (RCR_RESET_MBFLAGS, AMCC_OP_REG_RCR);
    amcc_write_opreg (RCR_RESET_ADDON, AMCC_OP_REG_RCR);
    for (j=0; j<500; j++)udelay(1000);
    amcc_write_opreg ((u32) 0, AMCC_OP_REG_RCR);

    amcc_write_opreg ((u32) 0, AMCC_OP_REG_ICSR);
    amcc_write_opreg ((u32) 0x00000000, AMCC_OP_REG_PTCR);

    /* some debug info */
    /* Power up: ICSR 0x00000C0C RCR 0x00000000 PTCR 0x80808080 */

    printk (KERN_DEBUG "amcc_init: ICSR: 0x%08x\n",
            amcc_read_opreg (AMCC_OP_REG_ICSR));
    printk (KERN_DEBUG "amcc_init:  RCR: 0x%08x\n",
            amcc_read_opreg (AMCC_OP_REG_RCR));
    printk (KERN_DEBUG "amcc_init: MBEF: 0x%08x\n",
            amcc_read_opreg (AMCC_OP_REG_MBEF));
    printk (KERN_DEBUG "amcc_init: PTCR: 0x%08x\n",
            amcc_read_opreg (AMCC_OP_REG_PTCR));

    printk (KERN_DEBUG "amcc_init: function terminated successfully\n");

    return (0);
}

/*------------------------------------------------------------------------------
 *
 *  Device finalization.
 *  It selectively frees all the resources allocated by the driver.
 *
 *----------------------------------------------------------------------------*/

static void
amcc_free_dev ()
{
    extern struct amcc_s5920 amcc_dev;
    extern struct amcc_buffer p2a;
    extern struct amcc_buffer a2p;
    int count = 0L;

    /* reset the ICSR register */
/*  amcc_write_opreg ((u32) 0, AMCC_OP_REG_ICSR); */

    if (amcc_dev.major)
        unregister_chrdev (amcc_dev.major, AMCC_NAME);

    if (amcc_dev.op_regs) {
        if (amcc_dev.op_regs_type == PCI_BASE_ADDRESS_SPACE_IO)
            release_region (amcc_dev.op_regs, AMCC_OP_REG_SIZE);
        else
            iounmap ((void *)amcc_dev.op_regs);		/* gsb added (void *) */
    }

    /* Free IRQ */
    if (amcc_dev.irq)
        free_irq (amcc_dev.irq, NULL);

    /* Free pass-thru regions -- S5920 has PT[1] thru PT[4] */
    for (count = 1; count <= 4; count++){
        if ((amcc_dev.pts[count]).virt_addr != NULL)
            iounmap ((amcc_dev.pts[count]).virt_addr);
    }

    /* Free Pass-Thru + FIFO buffers */
    if (p2a.buf)
        kfree (p2a.buf);

    if (a2p.buf)
        kfree (a2p.buf);
}


/*------------------------------------------------------------------------------
 *
 *  Interrupt handler.
 *
 *----------------------------------------------------------------------------*/
/*
static void
amcc_irq_handler (int irq, void *dev, struct pt_regs *regs)
{
    extern volatile u32 icsr;
    extern struct amcc_timeinfo amcc_ti;

    do_gettimeofday (&(amcc_time.irq));

    if (irq != amcc_dev.irq) {
        printk (KERN_WARNING "amcc: irq: spurious interrupt!\n");
        return;
    }

    icsr = amcc_read_opreg (AMCC_OP_REG_ICSR);
    printk (KERN_DEBUG "amcc: got irq: ICSR = 0x%08lx\n", icsr);

    if (!(icsr & ICSR_INT_ASSERTED))
        return;

}
*/
